/**
Copyright Philipp Tölke, 2006-2007

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License version 2 as published by the Free
Software Foundation.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
St, Fifth Floor, Boston, MA  02110-1301  USA
**/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
extern "C" {
#include <jpeglib.h>
}
#include <limits.h>

#include <math.h>

#include "blibfinder.h"

#define MAX_CPP 3

static void avg (JSAMPLE pixels[], JSAMPLE **result, int cpp, int n);
static double abstand(JSAMPLE a[], JSAMPLE b[], int cpp);

struct clusterlist { //{{{
    JSAMPLE* pixel;
    JSAMPLE *avg;
    unsigned int count;
};
//}}}
static clusterlist* clusterArray = NULL;
static unsigned int clusterArrayI = 0;
static inline void initArray(int n) { //{{{
    if (clusterArray == NULL) {
        clusterArray = new clusterlist[n];
    }
    clusterArrayI = 0;
}
//}}}
static inline void resetArray() { //{{{
    for (unsigned int i = 0; i < clusterArrayI; i++) {
        delete[] clusterArray[i].pixel;
        free(clusterArray[i].avg);
    }
}
//}}}
static void getClusterList(const JSAMPROW pixels, const int cpp, const int n, const float epsilon) { //{{{
    for (int index = 0; index < n*cpp; index += cpp) {
        boolean found = false;

        for (unsigned int i = 0; i < clusterArrayI; i++) {
            double d = abstand(pixels+index, clusterArray[i].avg, cpp);
            if (d < epsilon) {
                found = true;

                for (int c = 0 ; c < cpp; c++)
                    clusterArray[i].pixel[cpp*(clusterArray[i].count)+c] = pixels[index+c];

                clusterArray[i].count++;

                avg(pixels+index, &(clusterArray[i].avg), cpp, clusterArray[i].count);
                break;
            }
        }

        if (!found) {
            clusterArray[clusterArrayI].count = 1;

            clusterArray[clusterArrayI].pixel = new JSAMPLE[n*cpp];

            for (int c = 0 ; c < cpp; c++)
                clusterArray[clusterArrayI].pixel[c] = pixels[index+c];

            clusterArray[clusterArrayI].avg = (JSAMPLE*)malloc(cpp*sizeof(JSAMPLE));
            memset(clusterArray[clusterArrayI].avg, 0, cpp*sizeof(JSAMPLE));

            avg(pixels+index, &(clusterArray[clusterArrayI].avg), cpp, 1);
            clusterArrayI++;

        }
    }
}
//}}}
static unsigned int getMaxList(clusterlist **maxCluster) { // {{{
    unsigned int maxLen = 0;
    for (unsigned int i = 0; i < clusterArrayI; i++) {
        if (clusterArray[i].count > maxLen) {
            maxLen = clusterArray[i].count;
            *maxCluster = clusterArray+i;
        }
    }
    return maxLen;
}
//}}}
/**
 * Removes a blib.
 *
 * Returns how many cpp-tupels there are in "ausgabe"
 */
unsigned int removeBlib (JSAMPROW pixels, JSAMPROW *ausgabe, int cpp, int n, float epsilon) { //{{{
    initArray(n);

    getClusterList(pixels, cpp, n, epsilon);

    clusterlist *maxCluster = NULL;

    unsigned int maxLen = getMaxList(&maxCluster);

    if (maxLen == 0)
        return 0;

    *ausgabe = new JSAMPLE[maxLen*cpp];
    int index = 0;
    for (unsigned int pi = 0; pi < maxCluster->count*cpp; pi++) {
        (*ausgabe)[index++] = maxCluster->pixel[pi];
    }

    resetArray();

    return maxLen;
}
//}}}
/**
 * Returns only the Blib (the Cluster with the greatest distance to the biggest Cluster)
 *
 * Returns how many cpp-tupels there are in "ausgabe"
 */
unsigned int findBlib (JSAMPROW pixels, JSAMPROW *ausgabe, int cpp, int n, float epsilon) { //{{{

    initArray(n);

    getClusterList(pixels, cpp, n, epsilon);

    clusterlist *maxCluster = NULL;

    unsigned int maxLen = getMaxList(&maxCluster);

    if (maxLen == 0)
        return 0;

    double maxAbstand = 0;
    clusterlist* minCluster = NULL;
    for (unsigned int i = 0; i < clusterArrayI; i++) {
        double d = abstand(clusterArray[i].pixel, maxCluster->pixel, cpp);
        if (d >= maxAbstand) {
            minCluster = clusterArray+i;
            maxAbstand = d;
        }
    }
    int minLen = minCluster->count;

    *ausgabe = new JSAMPLE[minLen*cpp];
    int index = 0;
    for (unsigned int pi = 0; pi < minCluster->count*cpp; pi++) {
        (*ausgabe)[index++] = minCluster->pixel[pi];
    }

    resetArray();

    return minLen;
}
//}}}
/**
 * Averages every cpp-th Value in pixels. Returns a float-Array of cpp floats in result
 *
 *
 * cpp: colors per pixel
 */
static void avg (JSAMPLE pixels[], JSAMPLE **result, int cpp, int n) { //{{{
    for (int i = 0; i < cpp; i++)
        (*result)[i] = ((*result)[i]*(n-1) + pixels[i])/n;
}
//}}}
/**
 * Calculates the distance between two cpp-Tupels in three dimensions.
 */
static double abstand(JSAMPLE a[], JSAMPLE b[], int cpp){ //{{{
    unsigned long ausgabe = 0;

    for (int i = 0; i < cpp; i++) {
        unsigned long t = a[i] - b[i];
        ausgabe += t*t;
    }

    return sqrt(ausgabe);
}
//}}}
#ifdef BLIBFINDERTEST
int main() { // {{{
    JSAMPLE test[] = {3,4,6,4,5,7};
    JSAMPLE *ausgabe;
    int x = removeBlib(test, &ausgabe, 3, 2, 10);
    delete[] ausgabe;
}
// }}}
#endif
