/**
Copyright Philipp Tölke, 2006-2007 {{{

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License version 2 as published by the Free
Software Foundation.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
St, Fifth Floor, Boston, MA  02110-1301  USA
}}} **/

#include <stdio.h>
#include <unistd.h>
extern "C" {
#include <jpeglib.h>
}
#include <string.h>

#include "blibfinder.h"
#include "pathhandler.h"
#include "outfile.h"
#include "jpegoutfile.h"
#include "tiffoutfile.h"

#include "error.h"

//output file types:
#define JPEG 0
#define TIFF 1

//for gettext:

#define _(Text) Text

boolean quiet = false;

/**
 * Takes the scanlines from nop pictures and returns a rearranged Array:
 *
 * scanlines:
 * Picture 1: >abcdefg
 *         2: >1234567
 *         3: >ABCDEFG
 *
 * ausgabe:
 * >a1A
 * >b2B
 * >c3C
 * >d4D
 * >e5E
 * >f6F
 * >g7G
 *
 * nop: number of pictures
 * cpp: colors per pixel
 */
static JSAMPARRAY pixelize(JSAMPROW scanlines[], int nop, int cpp, int width) { //{{{
    JSAMPARRAY ausgabe = new JSAMPLE*[width];
    for (int i = 0; i < width; i++)
        ausgabe[i] = new JSAMPLE[nop*cpp];

    for (int sp = 0; sp < width; sp++)    //Column (=Pixel-Coulums)
        for (int z = 0; z < nop; z++)     //Row  (=Picture)
            for (int c = 0; c < cpp; c++) //Color
                ausgabe[sp][cpp*z+c] = scanlines[z][cpp*sp+c];

    return ausgabe;
}
//}}}
static int process_files(char* filename[], const char* outfilename, int n, Outfile* Out, unsigned int (blibfunction)(JSAMPROW, JSAMPROW*, int, int, float))  { //{{{
    int error;
    struct jpeg_decompress_struct *cinfos = new jpeg_decompress_struct[n];

    struct jpeg_error_mgr jerr;


    FILE** infiles = new FILE*[n];
    JSAMPROW out_row[1];

    JSAMPARRAY buffers = new JSAMPROW[n];

    unsigned int width = 0;
    int cpp = 0;

    for (int f = 0; f < n; f++) {
        if ((infiles[f] = fopen(filename[f], "rb")) == NULL) {
            fprintf(stderr, _("can't open %s\n"), filename[f]);
            for ( int f_ = 0; f_ < f; f_++)
                fclose(infiles[f_]);
            delete[] cinfos;
            delete[] infiles;
            delete[] buffers;
            return INFILENOTFOUND;
        }
    }

    for (int f = 0; f < n; f++) {
        (cinfos+f)->err = jpeg_std_error(&jerr);
        jpeg_create_decompress(cinfos+f);
        jpeg_stdio_src(cinfos+f, infiles[f]);
        error = jpeg_read_header(cinfos+f, TRUE);
        if (error != 1) printf(_("Fehler beim Einlesen der JPEG-Header von Datei %s.\n"), filename[f]);
        error = jpeg_start_decompress(cinfos+f);
        if (error != 1) printf(_("Fehler beim Start der Dekomprimierung von Datei %s.\n"), filename[f]);

        if ( f == 0 ) {
            cpp = cinfos->output_components;
            width = cinfos->output_width;
        } else {
            if (cinfos[f].output_components != cpp || cinfos[f].output_width != width) {
                for ( int f_ = 0; f_ < f; f_++) {
                    error = jpeg_finish_decompress(cinfos+f_);
                    if (error != 1) printf(_("Fehler beim Beenden der Dekomprimierung von Datei %s.\n"), filename[f]);
                    jpeg_destroy_decompress(cinfos+f_);
                    fclose(infiles[f_]);

                    delete[] buffers[f_];
                }
                delete[] cinfos;
                delete[] infiles;
                delete[] buffers;
                return INPUTNOTUNIFORM;
            }
        }

        buffers[f] = new JSAMPLE[cpp*width];
    }

    if (Out->initialize(outfilename, &jerr, cinfos[0].output_width, cinfos[0].output_height, cinfos[0].output_components, cinfos[0].out_color_space)==OUTFILENOTFOUND) {
        fprintf(stderr, _("can't open %s\n"), outfilename);

        for ( int f = 0; f < n; f++) {
            error = jpeg_finish_decompress(cinfos+f);
            if (error != 1) printf(_("Fehler beim Beenden der Dekomprimierung von Datei %s.\n"), filename[f]);
            jpeg_destroy_decompress(cinfos+f);
            fclose(infiles[f]);

            delete[] buffers[f];
        }
        delete[] cinfos;
        delete[] infiles;
        delete[] buffers;
        return OUTFILENOTFOUND;
    }

    out_row[0] = new JSAMPLE[cpp*width];

    while (cinfos->output_scanline < cinfos->output_height) {
        if (!quiet && cinfos->output_scanline % 10 == 0) fprintf(stdout, "%d/%d\n", cinfos->output_scanline, cinfos->output_height);
        for (int f = 0; f < n; f++) {
            error = jpeg_read_scanlines(cinfos+f, buffers+f, 1);
            if (error != 1) printf(_("Fehler beim Lesen aus Datei %s.\n"), filename[f]);
        }
        JSAMPARRAY pixellines = pixelize(buffers, n, cpp, width);
        for (unsigned int p = 0; p < width; p++) {
            JSAMPROW blibfree;
            /*int remaining_pixels = */ blibfunction(pixellines[p], &blibfree, cpp, n, 10);
            for ( int c = 0; c < cpp; c++) {
                out_row[0][cpp*p+c] = blibfree[c];
            }
            delete[] pixellines[p];
            delete[] blibfree;
        }
        delete[] pixellines;
        Out->write_scanline(out_row);
    }

    delete[] out_row[0];

    for ( int f = 0; f < n; f++) {
        error = jpeg_finish_decompress(cinfos+f);
        if (error != 1) printf(_("Fehler beim Beenden der Dekomprimierung von Datei %s.\n"), filename[f]);
        jpeg_destroy_decompress(cinfos+f);
        fclose(infiles[f]);

        delete[] buffers[f];
    }

    Out->destroy();

    delete[] buffers;
    delete[] infiles;
    delete[] cinfos;

    return 0;
}
//}}}
static void printUsage(int optopt, char* argv[]) { //{{{
    if (optopt != -1) {
        printf(_("Ungueltiges Argument: %c\n"), optopt);
    }

    printf(_("Bitte aufrufen mit %s [-q] [-k] [-t] [-o <pfad>] (<dateiname>)*.\n"), argv[0]);
    printf(_("Also z. B. %s 1.jpg 2.jpg 3.jpg.\n"), argv[0]);
    printf(_("\nBedeutung der Argumente:\n"));
    printf(_("\t-q\tNicht alle 10 Zeilen die aktuelle Zeile ausgeben\n"));
    printf(_("\t-k\t'Ausreisser' behalten und nicht loeschen\n"));
    printf(_("\t-t\tAusgabe als tiff\n"));
    printf(_("\t-o\tAusgabeort angeben:\n"));
    printf(_("\t\tIst 'pfad' eine Datei, wird diese Datei als Ausgabe verwendet\n"));
    printf(_("\t\tIst 'pfad' ein Verzeichnis, wird in dieses Verzeichnis die\n"));
    printf(_("\t\tAusgabe geschrieben\n"));
#ifdef _WIN32
    printf(_("\nBitte Enter druecken.\n"));
    getc(stdin);
#endif
}
//}}}
int main(int argc, char* argv[]) { //{{{
    char* outfilename = "out.jpg";
    char* outfilepath = "";
    int filetype = JPEG;

    opterr = 0;

    unsigned int (*blibfunction)(JSAMPROW, JSAMPROW*, int, int, float) = &removeBlib;

    int c;
    while ((c = getopt (argc, argv, "khto:q")) != -1)
        switch (c) {
            case 't':
                filetype = TIFF;
                outfilename="out.tiff";
                break;
            case 'o':
                outfilepath = optarg;
                break;
            case 'h':
                printUsage(-1, argv);
                return 0;
                break;
            case 'k':
                blibfunction = &findBlib;
                break;
            case 'q':
                quiet = true;
                break;
            case '?':
                printUsage(optopt, argv);
                return 1;
                break;
        }

    if (argc == optind) { // No arguments but t
        printUsage(-1, argv);
        return 1;
    }

    const char * outfile = getFullOutFile(argv[optind], outfilename, outfilepath);

    if (!quiet) {
        printf(_("%d Eingabe-Dateien, "), argc-optind);
        switch (filetype) {
            case JPEG:
                printf(_("Ausgabe als jpeg\n"));
                break;
            case TIFF:
                printf(_("Ausgabe als tiff\n"));
                break;
        }
        printf(_("Ausgabe in die Datei: %s\n"), outfile);
    }

    Outfile* o;
    switch (filetype) {
        case JPEG:
            o = new JpegOutfile();
            break;
        case TIFF:
            o = new TIFFOutfile();
            break;
        default:
           return 1;
    }

    int x = process_files(&(argv[optind]), outfile, argc-optind, o, blibfunction);

    switch (x) {
        case INFILENOTFOUND:
            printf(_("Mindestens eine Eingabedatei konnte nicht geöffnet werden!\n"));
            break;
        case OUTFILENOTFOUND:
            printf(_("Die Ausgabedatei konnte nicht geöffnet werden!\n"));
            break;
        case 0:
            printf(_("Alles OK.\n"));
            break;
    }

    delete o;

    delete [] outfile;

    return 0;
}
//}}}
